from pydantic import BaseModel, Field, Extra, root_validator
from .Entity import Permission, Group, Member, Friend
from .Message import MessageChain
from .Types import MessageType
from typing import Optional, Literal, Union, Type, Any
from datetime import datetime


class BaseEvent(BaseModel):
    type: str

    class Config:
        extra = Extra.allow

    def __str__(self):
        return f'[{str(self.json(ensure_ascii=False))}]'


class BotOnlineEvent(BaseEvent):
    type: Literal['BotOnlineEvent']
    qq: int

    def __str__(self):
        return f'[{self.type}: qq={self.qq}]'


class BotOfflineEventActive(BaseEvent):
    type: Literal['BotOfflineEventActive']
    qq: int

    def __str__(self):
        return f'[{self.type}: qq={self.qq}]'


class BotOfflineEventForce(BaseEvent):
    type: Literal['BotOfflineEventForce']
    qq: int

    def __str__(self):
        return f'[{self.type}: qq={self.qq}]'


class BotOfflineEventDropped(BaseEvent):
    type: Literal['BotOfflineEventDropped']
    qq: int

    def __str__(self):
        return f'[{self.type}: qq={self.qq}]'


class BotReloginEvent(BaseEvent):
    type: Literal['BotReloginEvent']
    qq: int

    def __str__(self):
        return f'[{self.type}: qq={self.qq}]'


class BotGroupPermissionChangeEvent(BaseEvent):
    type: Literal['BotGroupPermissionChangeEvent']
    origin: Permission
    current: Permission
    group: Group

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={repr(self.origin)}, new={repr(self.current)}, group={repr(self.group)}]'


class BotMuteEvent(BaseEvent):
    type: Literal['BotMuteEvent']
    durationSeconds: int
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: durationSeconds={self.durationSeconds}, operator={repr(self.operator)}]'


class BotUnmuteEvent(BaseEvent):
    type: Literal['BotUnmuteEvent']
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: operator={repr(self.operator)}]'


class BotJoinGroupEvent(BaseEvent):
    type: Literal['BotJoinGroupEvent']
    group: Group

    def __str__(self):
        return f'[{self.type}: group={repr(self.group)}]'


class BotLeaveEventActive(BaseEvent):
    type: Literal['BotLeaveEventActive']
    group: Group

    def __str__(self):
        return f'[{self.type}: group={repr(self.group)}]'


class BotLeaveEventKick(BaseEvent):
    type: Literal['BotLeaveEventKick']
    group: Group

    def __str__(self):
        return f'[{self.type}: group={repr(self.group)}]'


class GroupRecallEvent(BaseEvent):
    type: Literal['GroupRecallEvent']
    authorId: int
    messageId: int
    time: datetime
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: authorId={self.authorId}, messageId={self.messageId}, time={self.time},' \
               f' group={repr(self.group)}, operator={repr(self.operator)}]'


class FriendRecallEvent(BaseEvent):
    type: Literal['FriendRecallEvent']
    authorId: int
    messageId: int
    time: int
    operator: int


class GroupNameChangeEvent(BaseEvent):
    type: Literal['GroupNameChangeEvent']
    origin: str
    current: str
    group: Group
    operator: bool

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={self.operator}]'


class GroupEntranceAnnouncementChangeEvent(BaseEvent):
    type: Literal['GroupEntranceAnnouncementChangeEvent']
    origin: str
    current: str
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={repr(self.operator)}]'


class GroupMuteAllEvent(BaseEvent):
    type: Literal['GroupMuteAllEvent']
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={repr(self.operator)}]'


class GroupAllowAnonymousChatEvent(BaseEvent):
    type: Literal['GroupAllowAnonymousChatEvent']
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={repr(self.operator)}]'


class GroupAllowConfessTalkEvent(BaseEvent):
    type: Literal['GroupAllowAnonymousChatEvent']
    origin: bool
    current: bool
    group: Group
    isByBot: bool

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, isByBot={self.isByBot}]'


class GroupAllowMemberInviteEvent(BaseEvent):
    type: Literal['GroupAllowMemberInviteEvent']
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={repr(self.operator)}]'


class MemberJoinEvent(BaseEvent):
    type: Literal['MemberJoinEvent']
    member: Member

    def __str__(self):
        return f'[{self.type}: ' \
               f'member={repr(self.member)}]'


class MemberLeaveEventKick(BaseEvent):
    type: Literal['MemberLeaveEventKick']
    member: Member
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: ' \
               f'member={repr(self.member)}, operator={repr(self.operator)}]'


class MemberLeaveEventQuit(BaseEvent):
    type: Literal['MemberLeaveEventQuit']
    member: Member

    def __str__(self):
        return f'[{self.type}: ' \
               f'member={repr(self.member)}]'


class MemberCardChangeEvent(BaseEvent):
    type: Literal['MemberCardChangeEvent']
    origin: str
    current: str
    member: Member
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, member={repr(self.member)}, operator={repr(self.operator)}]'


class MemberSpecialTitleChangeEvent(BaseEvent):
    type: Literal['MemberSpecialTitleChangeEvent']
    origin: str
    current: str
    member: Member

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, member={repr(self.member)}]'


class MemberPermissionChangeEvent(BaseEvent):
    type: Literal['MemberPermissionChangeEvent']
    origin: str
    current: str
    member: Member

    def __str__(self):
        return f'[{self.type}: ' \
               f'origin={self.origin}, new={self.current}, member={repr(self.member)}]'


class MemberMuteEvent(BaseEvent):
    type: Literal['MemberMuteEvent']
    durationSeconds: int
    member: Member
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: ' \
               f'durationSeconds={self.durationSeconds}, member={repr(self.member)}, operator={repr(self.operator)}]'


class MemberUnmuteEvent(BaseEvent):
    type: Literal['MemberUnmuteEvent']
    member: Member
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type}: ' \
               f'member={repr(self.member)}, operator={repr(self.operator)}]'


class Message(BaseEvent):
    # fusion class
    type: MessageType
    messageChain: MessageChain
    sender: Union[Friend, Member]

    def __init__(self, messageChain: MessageChain = None, **kwargs):
        if messageChain:
            messageChain = MessageChain.parse_obj(messageChain)
        super().__init__(messageChain=messageChain, **kwargs)

    @property
    def member(self):
        if isinstance(self.sender, Member):
            return self.sender
        return None

    @property
    def friend(self):
        if isinstance(self.sender, Friend):
            return self.sender
        return None


class NewFriendRequestEvent(BaseEvent):
    type: Literal['NewFriendRequestEvent']
    requestId: int = Field(..., alias="eventId")
    supplicant: int = Field(..., alias="fromId")  # 即请求方 QQ
    sourceGroup: Optional[int] = Field(..., alias="groupId")
    nickname: str = Field(..., alias="nick")


class MemberJoinRequestEvent(BaseEvent):
    type: Literal['MemberJoinRequestEvent']
    requestId: int = Field(..., alias="eventId")
    supplicant: int = Field(..., alias="fromId")  # 即请求方 QQ
    sourceGroup: Optional[int] = Field(..., alias="groupId")
    groupName: str = Field(..., alias="groupName")
    nickname: str = Field(..., alias="nick")


def events() -> Type:
    event_types = [
        Message,

        BotOnlineEvent,
        BotOfflineEventActive,
        BotOfflineEventForce,
        BotOfflineEventDropped,
        BotReloginEvent,
        BotGroupPermissionChangeEvent,
        BotMuteEvent,
        BotUnmuteEvent,
        BotJoinGroupEvent,
        BotLeaveEventActive,
        BotLeaveEventKick,
        GroupNameChangeEvent,
        GroupEntranceAnnouncementChangeEvent,
        GroupMuteAllEvent,
        GroupRecallEvent,
        FriendRecallEvent,
        GroupAllowAnonymousChatEvent,
        GroupAllowConfessTalkEvent,
        GroupAllowMemberInviteEvent,
        MemberJoinEvent,
        MemberLeaveEventKick,
        MemberLeaveEventQuit,
        MemberCardChangeEvent,
        MemberSpecialTitleChangeEvent,
        MemberPermissionChangeEvent,
        MemberMuteEvent,
        MemberUnmuteEvent,
        NewFriendRequestEvent,
        MemberJoinRequestEvent,

        BaseEvent
    ]
    return Union[tuple(event_types)]


Events = events()



from enum import Enum


class NewFriendRequestResponse(Enum):
    ACCEPT = 0
    REFUSE = 1
    REFUSE_AND_BLACKLIST = 2


class MemberJoinRequestResponse(Enum):
    ACCEPT = 0
    REFUSE = 1
    IGNORE = 2
    REFUSE_AND_BLACKLIST = 3
    IGNORE_AND_BLACKLIST = 4


class MessageType(str, Enum):
    GROUP = 'GroupMessage'
    FRIEND = 'FriendMessage'
    TEMP = 'TempMessage'



from abc import abstractmethod
from enum import Enum
from typing import List, Dict, Optional, overload, Iterable, Union, Literal, Type, Any
from pydantic import Field, validator, HttpUrl, BaseModel, Extra, root_validator
from .Constant import qq_emoji_text_list
import datetime
from collections import MutableSequence
import json

__all__ = [
    "BaseMessageComponent",
    'Plain',
    'Source',
    'At',
    'AtAll',
    'Face',
    'Image',
    'FlashImage',
    'Quote',
    'App',
    'Json',
    'Xml',
    'Poke',
    'MessageChain',
    'BotMessage'
]


class BaseMessageComponent(BaseModel):
    type: str

    class Config:
        extra = Extra.allow

    def __str__(self):
        """
        for plain text extraction
        :return: human readable text component
        """
        return 'Unknown message component: ' + json.dumps(self.dict(), ensure_ascii=False, indent=4)

    def __repr__(self):
        return '[Unknown]'


class Source(BaseMessageComponent):
    """
    The source of the message
    Not a valid component for outbound message
    """
    type: Literal['Source'] = 'Source'
    id: int  # the sequence id
    time: datetime.datetime  # the timestamp the message was sent

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Source: id={self.id}, time={self.time}]'


class Plain(BaseMessageComponent):
    """
    Text component
    Available for outbound message
    """
    type: Literal['Plain'] = 'Plain'
    text: str  # text content

    def __init__(self, text: str, **kwargs):
        """
        Construct text component

        :param text: message text
        """
        super().__init__(text=text, **kwargs)

    def __str__(self):
        return self.text

    def __repr__(self):
        return f'[Plain: {self.text}]'


class At(BaseMessageComponent):
    """
    At component
    Available for outbound message
    """
    type: Literal['At'] = 'At'
    target: int  # target qq number
    display: str  # display name (only used for inbound at), '@' already included

    def __init__(self, target, **kwargs):
        """
        Construct at component

        :param target: target qq number
        """
        super().__init__(target=target, **kwargs)

    def __str__(self):
        return self.display

    def __repr__(self):
        return f'[At: target={self.target}, display={self.display}]'


class AtAll(BaseMessageComponent):
    """
    AtAll component
    Available for outbound message
    """
    type: Literal['AtAll'] = 'AtAll'

    def __str__(self):
        return '@All'

    def __repr__(self):
        return f'[AtAll]'


class Face(BaseMessageComponent):
    """
    Face component
    Available for outbound message
    """
    type: Literal['Face'] = 'Face'
    faceId: int  # face id, see qq_emoji_text_list for details

    def __init__(self, faceId, **kwargs):
        """
        Construct Face component

        :param face_id: unsigned 8-bit face id
        """
        super().__init__(faceId=faceId, **kwargs)

    def __str__(self):
        return qq_emoji_text_list[self.faceId]

    def __repr__(self):
        return f'[Face: id={self.faceId}, {qq_emoji_text_list[self.faceId]}]'


class Image(BaseMessageComponent):
    """
    Received image component
    Available for outbound message
    """
    type: Literal['Image'] = 'Image'
    imageId: Optional[str]
    url: Optional[HttpUrl]
    path: Optional[str]

    def __init__(self,
                 imageId: Optional[str] = None,
                 url: Optional[HttpUrl] = None,
                 path: Optional[str] = None,
                 **kwargs):
        """
        Construct Image from mirai styled uuid, http url or local file path

        :param imageId: uuid as str (see https://github.com/mamoe/mirai-api-http/blob/master/MessageType.md#image)
        :param url: image url (http url)
        :param path: local image path (absolute path or relative path under "plugins/MiraiAPIHTTP/images")
        """
        super().__init__(imageId=imageId, url=url, path=path, **kwargs)

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Image: {self.imageId}]'

    @property
    def image_type(self):
        """
        get image type from pattern

        :return: 'friend' or 'group', if uuid is invalid format, return 'unknown'
        """
        if self.imageId.startswith('/'):
            return 'friend'
        elif self.imageId.startswith('{'):
            return 'group'
        else:
            'unknown'


class FlashImage(BaseMessageComponent):
    """
    Received image component
    Available for outbound message
    """
    type: Literal['FlashImage'] = 'FlashImage'
    imageId: Optional[str]
    url: Optional[HttpUrl]
    path: Optional[str]

    def __init__(self,
                 imageId: Optional[str] = None,
                 url: Optional[HttpUrl] = None,
                 path: Optional[str] = None,
                 **kwargs):
        """
        Construct Image from mirai styled uuid, http url or local file path

        :param imageId: uuid as str (see https://github.com/mamoe/mirai-api-http/blob/master/MessageType.md#image)
        :param url: image url (http url)
        :param path: local image path (absolute path or relative path under "plugins/MiraiAPIHTTP/images")
        """
        super().__init__(imageId=imageId, url=url, path=path, **kwargs)

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Image: {self.imageId}]'

    @property
    def image_type(self):
        """
        get image type from pattern

        :return: 'friend' or 'group', if uuid is invalid format, return 'unknown'
        """
        if self.imageId.startswith('/'):
            return 'friend'
        elif self.imageId.startswith('{'):
            return 'group'
        else:
            'unknown'


class Xml(BaseMessageComponent):
    """
    Xml component
    Available for outbound message
    """
    type: Literal['Xml'] = 'Xml'
    xml: str  # xml content

    def __init__(self, xml: str, **kwargs):
        """
        Construct Xml component

        :param xml: str contains xml
        """
        super().__init__(xml=xml, **kwargs)

    def __str__(self):
        return self.xml

    def __repr__(self):
        return f'[Xml: {self.xml}]'


class Json(BaseMessageComponent):
    """
    Json component
    Available for outbound message
    """
    type: Literal['Json'] = 'Json'
    Json: str = Field(..., alias="json")  # json content

    def __init__(self, Json: str, **kwargs):
        """
        Construct Json component

        :param json: json content
        """
        super().__init__(Json=Json, **kwargs)

    def __str__(self):
        return self.json

    def __repr__(self):
        return f'[Json: {self.json}]'


class App(BaseMessageComponent):
    """
    App component
    Available for outbound message
    """
    type: Literal['App'] = 'App'
    content: str  # app content

    def __init__(self, content: str, **kwargs):
        """
        Construct App component

        :param content: app content
        """
        super().__init__(content=content, **kwargs)

    def __str__(self):
        return str(self.content)

    def __repr__(self):
        return f'[App: {str(self.content)}]'


class PokeChoices(str, Enum):
    Poke = 'Poke'
    ShowLove = 'ShowLove'
    Like = 'Like'
    Heartbroken = 'Heartbroken'
    SixSixSix = 'SixSixSix'
    FangDaZhao = 'FangDaZhao'


class Poke(BaseMessageComponent):
    """
    Poke component
    Available for outbound message
    """
    type: Literal['Poke'] = 'Poke'
    name: PokeChoices

    def __init__(self, name: PokeChoices, **kwargs):
        """
        Construct Poke component

        :param name: Poke content
        """
        super().__init__(name=name, **kwargs)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'[Poke: {self.name}]'


class QuoteMessageChain(BaseModel):
    # stores the actual components
    __root__: List[Union[Source, Plain, Image, At, Face, FlashImage, AtAll, Xml, Json, App, Poke, BaseMessageComponent]]

    def insert(self, index: int, object) -> None:
        self.__root__.insert(index, object)

    def __setitem__(self, i: int, o) -> None:
        self.__root__.__setitem__(i, o)

    def __delitem__(self, i: int) -> None:
        self.__root__.__delitem__(i)

    def __add__(self, value):
        # merge two message chain or append one component
        if isinstance(value, BaseMessageComponent):
            self.__root__.append(value)
            return self
        elif isinstance(value, MessageChain):
            self.__root__ += value.__root__
            return self

    def __str__(self) -> str:
        return ''.join([str(i) for i in self.__root__])

    def __iter__(self):
        return self.__root__.__iter__()

    def __getitem__(self, index):
        return self.__root__.__getitem__(index)

    def has(self, component_class) -> bool:
        """
        test if any item in MessageChain is component_class

        :param component_class: the class for the component
        :return: boolean
        """
        for i in self:
            if isinstance(i, component_class):
                return True
        else:
            return False

    def __len__(self) -> int:
        return len(self.__root__)

    def get_first(self, component_class) -> Optional[BaseMessageComponent]:
        """
        Get the first component with component_class

        :param component_class: the class for the component
        :return: None or the component
        """
        for i in self.__root__:
            if isinstance(i, component_class):
                return i

    def get_all(self, component_class) -> List[BaseMessageComponent]:
        return [i for i in self.__root__ if isinstance(i, component_class)]

    def get_source(self) -> Source:
        result = self.get_first(Source)
        assert isinstance(result, Source)
        return result


class Quote(BaseMessageComponent):
    """
    Quote component
    Not a valid component for outbound message
    """
    type: Literal['Quote'] = 'Quote'
    id: int  # origin message id
    groupId: int  # origin group id
    senderId: int  # origin sender id
    targetId: int  # group id
    origin: QuoteMessageChain  # origin message (Source and other content without Quote)

    def __init__(self, origin=None, **kwargs):
        origin = QuoteMessageChain.parse_obj(origin)
        super().__init__(origin=origin, **kwargs)

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Quote: id={self.id}]'


class MessageChain(BaseModel):
    # stores the actual components
    __root__: List[
        Union[Source, Plain, Image, Quote, At, Face, FlashImage, AtAll, Xml, Json, App, Poke, BaseMessageComponent]]

    def insert(self, index: int, object) -> None:
        self.__root__.insert(index, object)

    def __setitem__(self, i: int, o) -> None:
        self.__root__.__setitem__(i, o)

    def __delitem__(self, i: int) -> None:
        self.__root__.__delitem__(i)

    def __add__(self, value):
        # merge two message chain or append one component
        if isinstance(value, BaseMessageComponent):
            self.__root__.append(value)
            return self
        elif isinstance(value, MessageChain):
            self.__root__ += value.__root__
            return self

    def __str__(self) -> str:
        return ''.join([str(i) for i in self.__root__])

    def __iter__(self):
        return self.__root__.__iter__()

    def __getitem__(self, index):
        return self.__root__.__getitem__(index)

    def has(self, component_class) -> bool:
        """
        test if any item in MessageChain is component_class

        :param component_class: the class for the component
        :return: boolean
        """
        for i in self:
            if isinstance(i, component_class):
                return True
        else:
            return False

    def __len__(self) -> int:
        return len(self.__root__)

    def get_first(self, component_class) -> Optional[BaseMessageComponent]:
        """
        Get the first component with component_class

        :param component_class: the class for the component
        :return: None or the component
        """
        for i in self.__root__:
            if isinstance(i, component_class):
                return i

    def get_all(self, component_class) -> List[BaseMessageComponent]:
        return [i for i in self.__root__ if isinstance(i, component_class)]

    def get_source(self) -> Source:
        result = self.get_first(Source)
        assert isinstance(result, Source)
        return result

    def get_quote(self) -> Quote:
        result = self.get_first(Quote)
        if result:
            assert isinstance(result, Quote)
            return result


class BotMessage(BaseModel):
    type: str = 'BotMessage'
    messageId: int
